# control.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Results database Control panel class.

Open and close databases and import and export data functions are available
on this panel.

"""

import tkinter
import os
import zipfile
import io
import email
import base64

try:
    import tnefparse
except ModuleNotFoundError:
    tnefparse = None

from solentware_base.minorbases.dbaseapi import dBaseapiError

from solentware_misc.workarounds import dialogues
from solentware_misc.api import utilities

from ..core.filespec import (ECFPLAYER_FILE_DEF, ECFCLUB_FILE_DEF)
from ..core import ecfdataimport
from ..core import ecfclubdb
from ..core import ecfplayerdb
from . import feedback
from . import control_lite
from . import tasklog


class Control(control_lite.Control):
    
    """The Control panel for a Results database.
    """
    
    _btn_ecfresultsfeedback = 103
    _btn_ecfmasterfile = 104
    _btn_ecfupdatefile = 105
    _btn_quitecfzippedfiles = 106
    _btn_copyecfmasterplayer = 107
    _btn_copyecfmasterclub = 108
    _btn_copyecfupdateplayer = 109
    _btn_copyecfupdateclub = 1001
    _btn_copyecfnewplayer = 1002

    def __init__(self, parent=None, cnf=dict(), **kargs):
        """Extend and define the results database control panel."""
        super(Control, self).__init__(
            parent=parent,
            cnf=cnf,
            **kargs)

        self.ecf_reference_file = None
        self._ecf_reference_widget = None
        self.ecfclubfile = None
        self.ecfplayerfile = None
        self.ecfclubupdatefile = None
        self.ecfplayerupdatefile = None
        self.ecfdatecontrol = None

    def close_update_resources(self):
        """Close any open files containing master data from ECF."""
        if self.ecfplayerfile != None:
            self.ecfplayerfile.close_context()
            self._delete_dbase_files(self.ecfplayerfile)
            self.ecfplayerfile = None
        if self.ecfclubfile != None:
            self.ecfclubfile.close_context()
            self._delete_dbase_files(self.ecfclubfile)
            self.ecfclubfile = None
        if self.ecfplayerupdatefile != None:
            self._delete_dbase_files(self.ecfplayerupdatefile)
            self.ecfplayerupdatefile.close_context()
            self.ecfplayerupdatefile = None
        if self.ecfclubupdatefile != None:
            self.ecfclubupdatefile.close_context()
            self._delete_dbase_files(self.ecfclubupdatefile)
            self.ecfclubupdatefile = None
        if self.ecfdatecontrol != None:
            self.ecfdatecontrol.destroy()
            self.ecfdatecontrol = None
            
    def describe_buttons(self):
        """Define all action buttons that may appear on Control page."""
        self.define_button(
            self._btn_closedatabase,
            text='Shut Database',
            tooltip='Close the open database.',
            underline=9,
            switchpanel=True,
            command=self.on_close_database)
        self.define_button(
            self._btn_ecfresultsfeedback,
            text='ECF Results Feedback',
            tooltip='Display a file containing feedback for a results submission to ECF.',
            underline=12,
            switchpanel=True,
            command=self.on_ecf_results_feedback)
        self.define_button(
            self._btn_importevents,
            text='Import Events',
            tooltip='Import event data exported by Export Events.',
            underline=0,
            switchpanel=True,
            command=self.on_import_events)
        self.define_button(
            self._btn_ecfmasterfile,
            text='ECF Master File',
            tooltip='Open a zipped ECF Master file.',
            underline=4,
            command=self.on_ecf_master_file)
        self.define_button(
            self._btn_ecfupdatefile,
            text='ECF Update File',
            tooltip='Open a zipped ECF Update file.',
            underline=4,
            command=self.on_ecf_update_file)
        self.define_button(
            self._btn_quitecfzippedfiles,
            text='Close File List',
            tooltip='Close the list of files in the zipped archive.',
            underline=1,
            command=self.on_quit_ecf_zipped_files)
        self.define_button(
            self._btn_copyecfmasterplayer,
            text='Show Master File (or New Players)',
            tooltip='Build new Master file for players.',
            underline=3,
            switchpanel=True,
            command=self.on_copy_ecf_master_player)
        self.define_button(
            self._btn_copyecfmasterclub,
            text='Show Master Club File',
            tooltip='Build new Master file for clubs.',
            underline=1,
            switchpanel=True,
            command=self.on_copy_ecf_master_club)
        self.define_button(
            self._btn_copyecfupdateplayer,
            text='Open Update Players',
            tooltip='Modify Master file for players on update file.',
            underline=0,
            switchpanel=True,
            command=self.on_copy_ecf_update_player)
        self.define_button(
            self._btn_copyecfupdateclub,
            text='Open Update Clubs',
            tooltip='Modify Master file for clubs on update file.',
            underline=0,
            switchpanel=True,
            command=self.on_copy_ecf_update_club)
        self.define_button(
            self._btn_copyecfnewplayer,
            text='Open New Players',
            tooltip='Build new Master file for players.',
            underline=1,
            switchpanel=True,
            command=self.on_copy_ecf_new_player)

    def on_ecf_results_feedback(self, event=None):
        """Do ECF feedback actions."""
        dlg = dialogues.askopenfilename(
            parent=self.get_widget(),
            title='Open ECF feedback email or attachment',
            #defaultextension='.txt',
            #filetypes=(('ECF feedback', '*.txt'),),
            initialdir='~')
        if not dlg:
            self.inhibit_context_switch(self._btn_ecfresultsfeedback)
            return
        try:
            feedbackfile = open(dlg, 'rb')
            try:
                self.get_appsys().set_kwargs_for_next_tabclass_call(
                    dict(datafile=(dlg, _get_feedback_text(feedbackfile))))
            finally:
                feedbackfile.close()
        except:
            dlg = dialogues.showinfo(
                master=self.get_widget(),
                message=''.join((
                    'File\n',
                    os.path.split(dlg)[-1],
                    '\ndoes not exist')),
                title=' '.join(['Open ECF feedback email or attachment']))
            return

    def on_quit_ecf_zipped_files(self, event=None):
        """Do quit import ECF Master File actions."""
        self._ecf_reference_widget.destroy()
        self._ecf_reference_widget = None
        self.datafilepath.configure(
            text=os.path.dirname(self.datafilepath.cget('text')))
        self.ecf_reference_file = None
        self.show_buttons_for_open_database()
        self.create_buttons()

    def on_copy_ecf_master_player(self, event=None):
        """Do copy ECF Master File (players) actions."""
        dbspec = self._get_memory_dBaseIII_from_zipfile(
            ecfplayerdb.ECFplayersDB)
        if dbspec is None:
            self.inhibit_context_switch(self._btn_copyecfmasterplayer)
            return
        self.get_appsys().set_kwargs_for_next_tabclass_call(dict(
            datafilespec=(dbspec[0], ecfplayerdb.PLAYERS, ecfplayerdb.PLAYERS),
            datafilename=dbspec[1],
            closecontexts=(ECFPLAYER_FILE_DEF,),
            tabtitle='Master List',
            copymethod=ecfdataimport.copy_ecf_players_post_2011_rules,
            ))

    def on_copy_ecf_master_club(self, event=None):
        """Do copy ECF club file actions."""
        dbspec = self._get_memory_dBaseIII_from_zipfile(
            ecfclubdb.ECFclubsDB)
        if dbspec is None:
            self.inhibit_context_switch(self._btn_copyecfmasterclub)
            return
        self.get_appsys().set_kwargs_for_next_tabclass_call(dict(
            datafilespec=(dbspec[0], ecfclubdb.CLUBS, ecfclubdb.CLUBS),
            datafilename=dbspec[1],
            closecontexts=(ECFCLUB_FILE_DEF,),
            tabtitle='Club List',
            copymethod=ecfdataimport.copy_ecf_clubs_post_2011_rules,
            ))
        
    def on_copy_ecf_update_player(self, event=None):
        """Do copy ECF player file update actions."""
        dbspec = self._get_memory_dBaseIII_from_zipfile(
            ecfplayerdb.ECFplayersUpdateDB)
        if dbspec is None:
            self.inhibit_context_switch(self._btn_copyecfupdateplayer)
            return
        self.get_appsys().set_kwargs_for_next_tabclass_call(dict(
            datafilespec=(dbspec[0], ecfplayerdb.PLAYERS, ecfplayerdb.PLAYERS),
            datafilename=dbspec[1],
            closecontexts=(ECFPLAYER_FILE_DEF,),
            tabtitle='Update Players',
            copymethod=ecfdataimport.update_ecf_players_post_2011_rules,
            ))

    def on_copy_ecf_update_club(self, event=None):
        """Do copy ECF club file update actions."""
        dbspec = self._get_memory_dBaseIII_from_zipfile(
            ecfclubdb.ECFclubsUpdateDB)
        if dbspec is None:
            self.inhibit_context_switch(self._btn_copyecfupdateclub)
            return
        self.get_appsys().set_kwargs_for_next_tabclass_call(dict(
            datafilespec=(dbspec[0], ecfclubdb.CLUBS, ecfclubdb.CLUBS),
            datafilename=dbspec[1],
            closecontexts=(ECFCLUB_FILE_DEF,),
            tabtitle='Update Clubs',
            copymethod=ecfdataimport.update_ecf_clubs_post_2011_rules,
            ))

    def on_copy_ecf_new_player(self, event=None):
        """Do copy ECF new player file actions."""
        dbspec = self._get_memory_dBaseIII_from_zipfile(
            ecfplayerdb.ECFplayersDB)
        if dbspec is None:
            self.inhibit_context_switch(self._btn_copyecfnewplayer)
            return
        self.get_appsys().set_kwargs_for_next_tabclass_call(dict(
            datafilespec=(dbspec[0], ecfplayerdb.PLAYERS, ecfplayerdb.PLAYERS),
            datafilename=dbspec[1],
            closecontexts=(ECFPLAYER_FILE_DEF,),
            tabtitle='New Players',
            copymethod=ecfdataimport.copy_ecf_players_post_2011_rules,
            ))

    def on_ecf_master_file(self, event=None):
        """Do unzip ECF master file actions"""
        if self.display_ecf_zipped_file_contents(
            self.datafilepath.cget('text'),
            ):
            self.show_buttons_for_ecf_master_file()
            self.create_buttons()

    def on_ecf_update_file(self, event=None):
        """Do unzip ECF Master File update actions."""
        if not dialogues.askyesno(
            master=self.get_widget(),
            message=''.join((
                'The ECF Update file is obsolescent.\n\n(These usually have ',
                'names like "ml70324u.zip" indicating date produced.)\n\n',
                'Click "Yes" to proceed.')),
            title='ECF Update File'):
            return None
        if self.display_ecf_zipped_file_contents(
            self.datafilepath.cget('text'),
            ):
            self.show_buttons_for_ecf_update_file()
            self.create_buttons()

    def display_ecf_zipped_file_contents(self, initialdir=''):
        """Display ECF master data with date for confirmation of update."""
        if not initialdir:
            initialdir = '~'
        dlg = dialogues.askopenfilename(
            parent=self.get_widget(),
            title='Open ECF data file',
            defaultextension='.zip',
            filetypes=(
                ('ECF master lists', '*.zip'),
                ),
            initialdir=initialdir)
        if not dlg:
            return

        ziparchive = zipfile.ZipFile(dlg, 'r')
        try:
            namelist = ziparchive.namelist()
            if len(namelist):
                frame = tkinter.Frame(master=self.get_widget())
                listbox = tkinter.Listbox(master=frame)
                yscrollbar = tkinter.Scrollbar(
                    master=frame,
                    orient=tkinter.VERTICAL,
                    command=listbox.yview)
                xscrollbar = tkinter.Scrollbar(
                    master=frame,
                    orient=tkinter.HORIZONTAL,
                    command=listbox.xview)
                listbox.configure(
                    yscrollcommand=yscrollbar.set,
                    xscrollcommand=xscrollbar.set)
                yscrollbar.pack(side=tkinter.RIGHT, fill=tkinter.Y)
                xscrollbar.pack(side=tkinter.BOTTOM, fill=tkinter.X)
                listbox.pack(fill=tkinter.BOTH, expand=tkinter.TRUE)
                frame.pack(fill=tkinter.BOTH, expand=tkinter.TRUE)
                listbox.insert(tkinter.END, *sorted(namelist))
                self.datafilepath.configure(text=dlg)
                self.ecf_reference_file = listbox
                self._ecf_reference_widget = frame
                return True
        finally:
            ziparchive.close()

    def open_file_from_ecf_zipped_master_file(
        self, dbdefinition, dbset, dbname, archive, element):
        """Display ECF master data with date for confirmation of update."""
        memory_file = None
        ziparchive = zipfile.ZipFile(archive, 'r')
        try:
            for za in ziparchive.namelist():
                if za == element:
                    memory_file = io.BytesIO(ziparchive.read(za))
                    break
        finally:
            ziparchive.close()

        ecffile = dbdefinition(memory_file)
        try:
            ecffile.open_context()
            return (ecffile, (archive, element))
        except dBaseapiError as msg:
            try:
                ecffile.close_context()
            except:
                pass
            dlg = dialogues.showinfo(
                master=self.get_widget(),
                message=str(msg),
                title=' '.join(['Open ECF player file']))
        except Exception as msg:
            try:
                ecffile.close_context()
            except:
                pass
            dlg = dialogues.showinfo(
                master=self.get_widget(),
                message=' '.join([str(Exception), str(msg)]),
                title=' '.join(['Open ECF player file']))

    def show_buttons_for_ecf_update_file(self):
        """Show buttons for actions allowed when importing new ECF club data."""
        self.hide_panel_buttons()
        self.show_panel_buttons(
            (self._btn_closedatabase,
             self._btn_quitecfzippedfiles,
             self._btn_copyecfupdateclub,
             self._btn_copyecfupdateplayer,
             self._btn_copyecfnewplayer))

    def show_buttons_for_ecf_master_file(self):
        """Show buttons for actions allowed importing new ECF player data."""
        self.hide_panel_buttons()
        self.show_panel_buttons(
            (self._btn_closedatabase,
             self._btn_quitecfzippedfiles,
             self._btn_copyecfmasterclub,
             self._btn_copyecfmasterplayer))

    def show_buttons_for_open_database(self):
        """Show buttons for actions allowed when database is open."""
        self.hide_panel_buttons()
        self.show_panel_buttons(
            (self._btn_closedatabase,
             self._btn_ecfresultsfeedback,
             self._btn_ecfmasterfile,
             self._btn_ecfupdatefile,
             self._btn_importevents))
    
    def _delete_dbase_files(self, dbaseobject):
        """Delete DBF files extracted from ECF master file ZIP file."""
        if not dbaseobject: return False

        for obj in dbaseobject.dBasefiles.values():
            if os.path.isfile(obj._file):
                try:
                    os.remove(obj._file)
                except:
                    pass

    def _get_memory_dBaseIII_from_zipfile(self, dbdefinition):
        """Create 'in-memory' dBaseIII file from zipped file."""
        selection = self.ecf_reference_file.curselection()
        if not selection:
            return
        selected_file = self.datafilepath.cget('text')
        selected_element = self.ecf_reference_file.get(selection)
        return self.open_file_from_ecf_zipped_master_file(
            dbdefinition,
            None,
            None,
            selected_file,
            selected_element,
            )


def _get_feedback_text(file):
    """Return feedback text from open file.

    Required text is assumed to be in 'text/plain' parts of message, where the
    'text/plain' may be inside an 'application/ms-tnef' attachment.

    """
    m = email.message_from_binary_file(file)

    # Assume file is a saved attachment file when there are no headers.
    if not m.keys():
        file.seek(0)
        return [line.rstrip() for line in file.readlines()]

    text = []
    for part in m.walk():
        ct = part.get_content_type()
        if ct == 'text/plain':
            text.extend(
                [line.rstrip()
                 for line in part.get_payload().encode('utf8').split(b'\n')])
        elif ct == 'application/ms-tnef':
            if not tnefparse:
                text.append(
                    b'Cannot process feedback: tnefparse is not installed.')
                continue

            # Assume the attachments are txt.
            # Feedback attachments have always been txt files: until December
            # 2016 these were not wrapped inside an application/ms-tnef
            # attachment.
            tnef = tnefparse.TNEF(base64.b64decode(part.get_payload()))
            for attachment in tnef.attachments:
                text.extend(
                    [line.rstrip()
                     for line in attachment.data.split(b'\n')])

    return text
